# Extracts polyspheres from Torsten Sillke's files
# Sorts them according to type and number
# Outputs a BurrTools file (simply gzip and rename as an xmpuzzle file)

import sys
from math import *

# First argument (if present)
# Number of spheres in these pieces (1-5)
grabindex = 3 # Default value
if len(sys.argv)>1: grabindex = int(sys.argv[1])

# Second argument (if present)
# Use long names including Sillke / Ishino numberings
long_names = 0
if len(sys.argv)>2: long_names = int(sys.argv[2])

# Third argument (if present)
# If set to 1 this will insert mirror image pieces
insert_mirrors = 1
if len(sys.argv)>3: insert_mirrors = int(sys.argv[3])

# Set to 0 normally
verbose = 0

# Do not change the variables below (unless you know what you're doing!)

# The ordering of the main categories
catorder = "SHXNM"

# Set input data file and column for sphere coordinates
if grabindex<6:
  data_file = "polyspheres.txt"
  col_sc = 19
else:
  data_file = "A"+str(grabindex)+"-sym.txt"
  col_sc = 0

# A flag indicating problems
problems = 0

# Compare floating point numbers x and y
# Returns 1 if x > y
# Returns -1 if x < y
# Returns 0 if x==y to within epsilon
def Compare(x,y,epsilon):
  if abs(x-y)<epsilon: return 0
  if x < y: return -1
  return 1

# Get the maximum distance between any pair of spheres
# Ignore this for now
def GetMaxDist(c,no):
  retval = 0.0
  for i in range(0,grabindex-1):
    for j in range(i+1,grabindex):
      dist = ((c[i][0]-c[j][0])**2+(c[i][1]-c[j][1])**2+(c[i][2]-c[j][2])**2)**0.5
      if dist>retval: retval=dist
  return retval

# Find the eigenvalues of the inertia tensor in decreasing magnitude
# Use the power method to find the largest eigenvalue
def GetEvals(it):
  xv = [1.0,1.0,1.0]
  xt = [0.0,0.0,0.0]
  for iter in range (0,1000):
    xmag = (xv[0]*xv[0]+xv[1]*xv[1]+xv[2]*xv[2])**0.5
    for i in range (0,3):
      xv[i] = xv[i]/xmag
    for i in range (0,3):
      xt[i] = it[i][0]*xv[0] + it[i][1]*xv[1] + it[i][2]*xv[2]
    for i in range (0,3):
      xv[i] = xt[i]

#  print "EV1: ",xt
  xmag = (xv[0]*xv[0]+xv[1]*xv[1]+xv[2]*xv[2])**0.5
  for i in range (0,3):
    xv[i] = xv[i]/xmag

  imax = 0
  for i in range (1,3):
    if abs(xv[i])>abs(xv[imax]): imax = i

  evals = [xt[imax]/xv[imax], 0.0, 0.0]

  g = -(it[0][0]+it[1][1]+it[2][2])
  h = it[0][0]*it[1][1] + it[0][0]*it[2][2] + it[1][1]*it[2][2] - it[0][1]**2 - it[0][2]**2 - it[1][2]**2
  k = -(it[0][0]*it[1][1]*it[2][2]+it[0][1]*it[1][2]*it[2][0]+it[1][0]*it[2][1]*it[0][2])
  k = k + it[2][0]*it[1][1]*it[0][2] + it[1][0]*it[0][1]*it[2][2] + it[0][0]*it[2][1]*it[1][2]
  b = g+evals[0]
  c = h+evals[0]*b
  disc = b*b-4.*c
  if disc<0: disc = 0.0
  r1 = 0.5*(-b+disc**0.5)
  r2 = 0.5*(-b-disc**0.5)
  if abs(r1)>abs(r2):
    evals[1] = r1
    evals[2] = r2
  else:
    evals[1] = r2
    evals[2] = r1
 
  test0 = k + evals[0]*(h + evals[0]*(g+evals[0]))
  test1 = k + evals[1]*(h + evals[1]*(g+evals[1]))
  test2 = k + evals[2]*(h + evals[2]*(g+evals[2]))
  eps = 0.0000001
  if abs(test0)>eps or abs(test1)>eps or abs(test2)>eps:
    print "ERROR??? test ",test0," ",test1," ",test2
    problems = 1

  return evals

# Alternative eigenvalue routine, from Wikipedia (seems not to give correct answers?)
def GetEvalsAlt(it):
  b = it[0][0] + it[1][1] + it[2][2]
  c = it[0][1]*it[1][0] + it[0][2]*it[2][0] +it[1][2]*it[2][1]
  c = c - (it[0][0]*it[1][1] + it[0][0]*it[2][2] + it[1][1]*it[2][2])
  d = it[0][0]*(it[1][1]*it[2][2] - it[1][2]*it[2][1])
  d = d + it[1][0]*(it[0][2]*it[2][1] - it[0][1]*it[2][2])
  d = d + it[2][0]*(it[0][1]*it[1][2] - it[0][2]*it[1][1])
 
  print b,c,d

  xx = (3.0*c - b*b)/3.0
  yy = (2.0*b*b*b - 9.0*b*c + 27.0*d)/27.0
  zz = yy*yy/4.0 + xx*xx*xx/27.0
  
  ii = sqrt(yy*yy/4.0 - zz)
  jj = -ii**(1./3.)
  kk = acos(-yy/(2.0*ii))
  mm = cos(kk/3.0)
  nn = sqrt(3.0)*sin(kk/3.0)
  pp = -b/3.0

  evals = [-2*jj*mm + pp, jj*(mm+nn) + pp, jj*(mm-nn) + pp]
  imax = 0
  for i in range(1,3):
    if evals[i]>evals[imax]: imax=i
  if i>0:
    tmp = evals[0]
    evals[0] = evals[imax]
    evals[imax] = tmp
  if evals[1]<evals[2]:
    tmp = evals[1]
    evals[1] = evals[2]
    evals[2] = tmp

  return evals

# Start of main
c = []
cg = [0.0,0.0,0.0]
xv = [0.0,0.0,0.0]
xt = [0.0,0.0,0.0]
coord_max = [0,0,0]
it = []
for i in range(grabindex):
  c.append([])
  for j in range(3):
    c[i].append(0)
for i in range(3):
  it.append([])
  for j in range(3):
    it[i].append(0)

p1 = 0
p2 = 0
p3 = 0

data = [] # An array for the polysphere data
ptrs = [] # Pointers for sorting
moms = [] # Moments of intertia
cats = [] # Categories

fi = open(data_file)
line = "#"
while line[0]=="#":
  line = fi.readline() # Skip all lines beginning with # (plus one more)

while 1:
  line = fi.readline()
  if not line: break
  if line[:3]=="end":
    break

  if int(line[4])==grabindex or grabindex>5:
    if grabindex<6:
      piece_name = line[:3]
      type = line[6]
    else:
      piece_name = str(p1+p2+1)
      checkok = 1
      type = ""
      typeloc = 46+(grabindex-6)*4
      test = line[typeloc:typeloc+3]
      if test==" S " or test=="HS ": type="S"
      if test=="H  ": type="H"
      if test=="H m": type="X"
      if test=="   ": type="N"
      if test=="  M": type="M"

    if type=="":
      print "Unclassified piece ",p1+p2," ???"
      problems = 3

    if verbose>0:
      print "Piece name is "+piece_name
      print "Piece coordinates are ",

    checkok = 1
    if type=="M": p2=p2+1
    else: p1=p1+1

    cg = [0.0,0.0,0.0] # Center of gravity
    for j in range (0,grabindex):
      baseindex = col_sc + j*4
      for i in range (0,3):
        c[j][i] = int(line[i+baseindex])
        cg[i] = cg[i]+c[j][i]
   
        if coord_max[i]<c[j][i]: coord_max[i] = c[j][i]

      if verbose>0:
        print "("+str(c[j][0])+","+str(c[j][1])+","+str(c[j][2])+")",
        if j<grabindex-1: print ",",
        else: print

    for i in range (0,3):
      cg[i] = cg[i]/float(grabindex)
      for j in range (0,3):
        it[i][j] = 0.0

    if verbose>0:
      print "CG: ("+str(cg[0])+","+str(cg[1])+","+str(cg[2])+")"

    for j in range (0,grabindex):
      it[0][0] = it[0][0] + (c[j][1]-cg[1])**2 + (c[j][2]-cg[2])**2
      it[0][1] = it[0][1] - (c[j][0]-cg[0])*(c[j][1]-cg[1])
      it[0][2] = it[0][2] - (c[j][0]-cg[0])*(c[j][2]-cg[2])

      it[1][0] = it[1][0] - (c[j][0]-cg[0])*(c[j][1]-cg[1])
      it[1][1] = it[1][1] + (c[j][0]-cg[0])**2 + (c[j][2]-cg[2])**2
      it[1][2] = it[1][2] - (c[j][1]-cg[1])*(c[j][2]-cg[2])

      it[2][0] = it[2][0] - (c[j][0]-cg[0])*(c[j][2]-cg[2])
      it[2][1] = it[2][1] - (c[j][1]-cg[1])*(c[j][2]-cg[2])
      it[2][2] = it[2][2] + (c[j][0]-cg[0])**2 + (c[j][1]-cg[1])**2

    if verbose>0:
      print it

    evals = GetEvals(it)
#    evals = GetEvalsAlt(it)

#    evals.append(GetMaxDist(c,grabindex))
    evals.append(p1+p2)

    if verbose>0:
      print "Eigenvalues:  ",evals[0]," ",evals[1]," ",evals[2]," maxdist ",evals[3]

    ptrs.append(len(data))
    data.append(line.rstrip())
    moms.append(evals)
    cats.append(catorder.find(type))

fi.close()

print "Piece counts: (normal ",p1,") (chiral ",p2,") (mirror images ",p3,")"
print "Total pieces: ",p1+p2+p3

if coord_max[0]<coord_max[1]: coord_max[0] = coord_max[1]
if coord_max[1]<coord_max[0]: coord_max[1] = coord_max[0]

# Simple sort by category and then moment of inertia

epsilon = 1.0e-6
for i in range (0,len(ptrs)-1):
  for j in range (i+1,len(ptrs)):
    swapem = 0
    test = Compare(cats[ptrs[j]], cats[ptrs[i]], epsilon)
    if test==-1: swapem = 1
    if test==0:
      test = Compare(moms[ptrs[j]][0], moms[ptrs[i]][0], epsilon)
      if test==1: swapem = 1
      if test==0:
        test = Compare(moms[ptrs[j]][1], moms[ptrs[i]][1], epsilon)
        if test==1: swapem = 1
        if test==0:
          test = Compare(moms[ptrs[j]][2], moms[ptrs[i]][2], epsilon)
          if test==1: swapem = 1
          if test==0:
            test = Compare(moms[ptrs[j]][3], moms[ptrs[i]][3], epsilon)
            if test==-1: swapem = 1

    if swapem==1:
      tmp = ptrs[i]
      ptrs[i] = ptrs[j]
      ptrs[j] = tmp


pieceno = 1
topsph = ""
topobl = ""
toppro = ""
eigtie = ""
lastname = ""

for j in range (0,len(ptrs)):
  strextra = ""
  if Compare(moms[ptrs[j]][1], moms[ptrs[j]][2], epsilon)==0:
    strextra = "SAUCER SYMMETRIC TOP"
  if Compare(moms[ptrs[j]][0], moms[ptrs[j]][1], epsilon)==0:
    strextra = "FOOTBALL SYMMETRIC TOP"
    if Compare(moms[ptrs[j]][1], moms[ptrs[j]][2], epsilon)==0:
      strextra = "SPHERICAL TOP"
  if j>0:
    if Compare(cats[ptrs[j-1]], cats[ptrs[j]], epsilon)==0:
      if Compare(moms[ptrs[j-1]][0], moms[ptrs[j]][0], epsilon)==0:
        if Compare(moms[ptrs[j-1]][1], moms[ptrs[j]][1], epsilon)==0:
          if Compare(moms[ptrs[j-1]][2], moms[ptrs[j]][2], epsilon)==0:
            strextra=" EIGENVALUE TIE"
            if Compare(moms[ptrs[j-1]][3], moms[ptrs[j]][3], epsilon)==0:
              strextra=" COMPLETE TIE"

  line = data[ptrs[j]]
  piecename = str(grabindex) + catorder[cats[ptrs[j]]] + str(pieceno)

  if strextra.find("SPHERICAL")>=0: topsph = topsph + piecename + ", "
  if strextra.find("SAUCER")>=0: topobl = topobl + piecename + ", "
  if strextra.find("FOOTBALL")>=0: toppro = toppro + piecename + ", "
  if strextra.find(" TIE")>=0 and grabindex<6: eigtie = eigtie + "(" + lastname + " and " +piecename + "), "
  lastname = piecename

  if long_names==1: piecename = piecename + "/" + str(ptrs[j]+1)
  if grabindex<6:
    print piecename,"  ",catorder[cats[ptrs[j]]],"  ",line,moms[ptrs[j]],strextra
  else:
    print piecename,"  ",catorder[cats[ptrs[j]]],"  ",line[:4*grabindex],moms[ptrs[j]],strextra

  pieceno = pieceno + 1
  if j+1<len(ptrs):
    if cats[ptrs[j+1]]!=cats[ptrs[j]]: pieceno = 1

fo = open("puzzleout"+str(grabindex)+".txt","w")

fo.write('<?xml version="1.0"?>\n')
fo.write('<puzzle version="2">\n')
fo.write(' <gridType type="2"/>\n') # Sphere grid
fo.write(' <colors/>\n')
fo.write(' <shapes>\n')

pieceno = 1
for j in range (0,len(ptrs)):
  line = data[ptrs[j]]
  piecename = str(grabindex) + catorder[cats[ptrs[j]]] + str(pieceno)
#  print "Piece name is '"+piecename+"'"

  torst = line[:3]
  torst = torst.strip()
#  print "Torsten number is '"+torst+"'"

  is1 = is2 = 0
  try:
    is1 = int(line[10:13])
  except:
    is1 = 0
  try:
    is2 = int(line[14:17])
  except:
    is2 = 0
#  print "Ishino numbers "+str(is1)+" and "+str(is2)

  if long_names==1:
    fullname = piecename+"/"+torst+"/"+str(is1)
    mirrorname = piecename.lower()+"/m"+torst+"/"+str(is2)
  else:
    fullname = piecename
    mirrorname = piecename.lower()

#  print "Full name is '"+fullname+"'"

  voxel_length = (coord_max[0]+1)*(coord_max[1]+1)*(coord_max[2]+1)

  voxelpat = ""
  for i in range(0,voxel_length):
    voxelpat = voxelpat + "_"
  voxelmir = ""
  for i in range(0,voxel_length):
    voxelmir = voxelmir + "_"

  newc = [0,0,0]
  for k in range (0,grabindex):
    baseindex = col_sc + k*4
    if k==0:
      test = 0
      for i in range (0,3):
        test = test+int(line[i+baseindex])
      if test%2==1: bumpup = [1,0,0]
      else: bumpup = [0,0,0]
    for i in range (0,3):
      newc[i] = int(line[i+baseindex])+bumpup[i]
#    print "("+str(newc[0])+","+str(newc[1])+","+str(newc[2])+")",
#    if k<grabindex-1: print ",",
#    else: print

# Note, first coordinate is layer (z)
# Second coordinate is x
# Third coordinate is y
#    voxcoor = 25*c[0]+5*c[2]+c[1]
    voxcoor = (coord_max[1]+1)*((coord_max[0]+1)*newc[0]+newc[2])+newc[1]
    voxelpat = voxelpat[0:voxcoor] + "#" + voxelpat[voxcoor+1:]

#    voxcoor = 25*c[0]+5*c[2]+(4-c[1])
    reflectx = coord_max[0]
    if reflectx%2==1: reflectx = coord_max[0]-1
    voxcoor = (coord_max[1]+1)*((coord_max[0]+1)*newc[0]+newc[2])+(reflectx-newc[1])
    voxelmir = voxelmir[0:voxcoor] + "#" + voxelmir[voxcoor+1:]

    if (newc[1]+newc[2])%2 <> newc[0]%2: checkok = 0

  if checkok==0:
    print "Check Failed!!"
    problems = 2

#  fo.write('  <voxel x="5" y="5" z="5" name="'+fullname+'" type="0">')
  fo.write('  <voxel x="'+str(coord_max[0]+1)+'" y="'+str(coord_max[1]+1)+'" z="'+str(coord_max[2]+1) )
  fo.write('" name="'+fullname+'" type="0">')
  fo.write(voxelpat)
  fo.write('</voxel>\n')

  if insert_mirrors==1 and catorder[cats[ptrs[j]]]=="M":
    fo.write('  <voxel x="'+str(coord_max[0]+1)+'" y="'+str(coord_max[1]+1)+'" z="'+str(coord_max[2]+1)+'" name="'+mirrorname+'" type="0">')
    fo.write(voxelmir)
    fo.write('</voxel>\n')
  
# On to the next piece

  pieceno = pieceno + 1
  if j+1<len(ptrs):
    if cats[ptrs[j+1]]!=cats[ptrs[j]]: pieceno = 1

#Finish off the puzzle file
fo.write(' </shapes>\n')
fo.write(' <problems/>\n')
#fo.write(' <comment/>\n')
fo.write('  <comment popup="">')
if long_names==1:
  fo.write('Piece notation is: "GB/Torsten/Ishino",\n')
  fo.write('  where "GB" is the notation of George Bell\n')
  fo.write('        "Torsten" is the notation of Torsten Sillke, http://www.mathematik.uni-bielefeld.de/~sillke/PENTA/notar\n')
  fo.write('        "Ishino" is the notation of Ishino Keiirchiro, http://www.asahi-net.or.jp/~rh5k-isn/Puzzle\n\n')
else:
  fo.write('Polyspheres labeled with George Bells notation.\n\n')

if len(eigtie)>0 and grabindex<6:
  fo.write('The following have identical three principal moments of inertia:\n')
  fo.write('  '+eigtie[0:len(eigtie)-2]+'\n')
if len(topsph)>0:
  fo.write('The following are spherical tops: '+topsph[0:len(topsph)-2]+'\n')
if len(toppro)>0:
  fo.write('The following are oblate (football-shaped) tops: '+toppro[0:len(toppro)-2]+'\n')
if len(topobl)>0:
  fo.write('The following are prolate (saucer-shaped) tops: '+topobl[0:len(topobl)-2]+'\n')

fo.write('</comment>\n')
fo.write('</puzzle>\n\n')
fo.close()
 
if problems>0:
  print "Finished, but with problems ", problems
else:
  print "Finished with no problems!!"
